from zope.interface import implements, alsoProvides, Interface
from zope.component import getMultiAdapter, provideUtility, provideAdapter
from souper.interfaces import ICatalogFactory
from souper.soup import get_soup, Record, NodeAttributeIndexer
from repoze.catalog.indexes.field import CatalogFieldIndex
from repoze.catalog.indexes.text import CatalogTextIndex
from repoze.catalog.indexes.keyword import CatalogKeywordIndex
try:
    from souper.plone.interfaces import ISoupRoot
    from souper.plone.locator import StorageLocator
except ImportError:
    from .interfaces import ISoupRoot
    from .locator import StorageLocator
    provideAdapter(StorageLocator, adapts=[Interface])

from rapido.core.interfaces import IStorage, IRapidoApplication

from .catalog import CatalogFactory
from .interfaces import IRecord


class SoupStorage(object):
    implements(IStorage)

    def __init__(self, context):
        self.context = context
        self.id = context.id
        self.root = self.context.root
        provideUtility(CatalogFactory(), ICatalogFactory, name=self.id)

    def initialize(self):
        """ setup the storage
        """
        alsoProvides(self.root, ISoupRoot)
        locator = StorageLocator(self.root)
        locator.storage(self.id)
        self._soup = get_soup(self.id, self.root)

    @property
    def soup(self):
        if not hasattr(self, '_soup'):
            self._soup = get_soup(self.id, self.root)
        return self._soup

    def create(self):
        """ return a new record
        """
        record = Record()
        rid = self.soup.add(record)
        return getMultiAdapter(
            (self.soup.get(rid), IRapidoApplication(self.context)),
            IRecord)

    def get(self, uid=None):
        """ return an existing record
        """
        record = self.soup.get(uid)
        if not record:
            return None
        return getMultiAdapter(
            (record, IRapidoApplication(self.context)),
            IRecord)

    def delete(self, record):
        """ delete a record
        """
        del self.soup[record.context]

    def search(self, query, sort_index=None, limit=None, sort_type=None,
            reverse=False, names=None, with_size=False):
        """ search for records
        """
        records = self.soup.lazy(query, sort_index=sort_index, limit=limit,
            sort_type=sort_type, reverse=reverse, names=names,
            with_size=with_size)
        db = IRapidoApplication(self.context)
        for record in records:
            yield getMultiAdapter((record(), db), IRecord)

    def records(self):
        for key in self.soup.data.keys():
            yield self.get(key)

    def rebuild(self):
        self.soup.rebuild()

    def clear(self):
        self.soup.clear()

    def reindex(self, record=None):
        if record:
            self.soup.reindex(records=[record.context])
        else:
            self.soup.reindex()

    @property
    def indexes(self):
        return self.soup.catalog.keys()

    def create_index(self, fieldname, indextype):
        catalog = self.soup.catalog
        field_indexer = NodeAttributeIndexer(fieldname)
        if indextype == 'field':
            catalog[fieldname] = CatalogFieldIndex(field_indexer)
        elif indextype == 'keyword':
            catalog[fieldname] = CatalogKeywordIndex(field_indexer)
        elif indextype == 'text':
            catalog[fieldname] = CatalogTextIndex(field_indexer)



from souper.interfaces import ICatalogFactory
from souper.soup import NodeAttributeIndexer
from zope.interface import implementer
from repoze.catalog.catalog import Catalog
from repoze.catalog.indexes.field import CatalogFieldIndex

@implementer(ICatalogFactory)
class CatalogFactory(object):

    def __call__(self, context=None):
        catalog = Catalog()
        id_indexer = NodeAttributeIndexer('id')
        catalog[u'id'] = CatalogFieldIndex(id_indexer)
        return catalog


from zope.interface import implementer
from zope.annotation import IAnnotations
from souper.interfaces import IStorageLocator
from souper.soup import SoupData

CACHE_PREFIX = 'soup_storage_%s'
SOUPPATHS = 'SOUPPATHS'
SOUPKEY = 'SOUP-%s'


@implementer(IStorageLocator)
class StorageLocator(object):

    def __init__(self, context):
        self.root = context

    def storage(self, sid):
        """return SoupData for the given soup id
        """
        context = self.traverse(self.path(sid))
        return self.soupdata(context, sid)

    def path(self, sid):
        """path to object with soupdata annotations for given soup id.
        relative to root.
        """
        paths = IAnnotations(self.root).get(SOUPPATHS, {})
        return paths.get(sid, '/')

    def traverse(self, path):
        """traverse to path relative to soups root and return the object there.
        """
        obj = self.root
        path = [_ for _ in path.split('/') if _]
        for name in path:
            try:
                obj = obj[name]
            except AttributeError:
                msg = u'Object at %s does not exist.' % '/'.join(path)
                raise ValueError(msg)
        return obj

    def soupdata(self, obj, sid):
        """fetches the soup data from objects annotations.
        """
        key = SOUPKEY % sid
        annotations = IAnnotations(obj)
        if key not in annotations:
            annotations[key] = SoupData()
        return annotations[key]



import unittest
import doctest
from pprint import pprint

optionflags = doctest.NORMALIZE_WHITESPACE | \
              doctest.ELLIPSIS | \
              doctest.REPORT_ONLY_FIRST_FAILURE

TESTFILES = [
    'soup.rst',
]


def test_suite():
    return unittest.TestSuite([
        doctest.DocFileSuite(
            filename,
            optionflags=optionflags,
            globs={
              'pprint': pprint,
            }
        ) for filename in TESTFILES
    ])

if __name__ == '__main__':                                 # pragma NO COVERAGE
    unittest.main(defaultTest='test_suite')                # pragma NO COVERAGE



from setuptools import setup, find_packages
import os

version = '1.0.1'

setup(name='rapido.souper',
    version=version,
    description="Souper persistence for Rapido",
    long_description=open("README.rst").read() + "\n" +
        open(os.path.join("docs", "HISTORY.txt")).read(),
    # Get more strings from
    # http://pypi.python.org/pypi?:action=list_classifiers
    classifiers=[
        "Programming Language :: Python",
    ],
    keywords='souper zodb rapido',
    author='Eric BREHAULT',
    author_email='ebrehault@gmail.com',
    url='https://github.com/plomino/rapido.souper',
    license='GPL',
    packages=find_packages(exclude=['ez_setup']),
    namespace_packages=['rapido'],
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'setuptools',
        'souper',
        'rapido.core',
    ],
    extras_require={
        'test': [
            'plone.app.testing',
        ],
    },
    entry_points="""
    # -*- Entry points: -*-
    """,
)



from zope.interface import implements
from copy import deepcopy

from rapido.core.interfaces import IRecordable

from .interfaces import IRecord


class Record(object):
    implements(IRecord, IRecordable)

    def __init__(self, context, app):
        self.context = context
        self.app = app

    def get(self, name, default=None):
        """ return an item value
        """
        if name in self:
            return self[name]
        else:
            return default

    def __getitem__(self, name):
        """ return an item value
        """
        return deepcopy(self.context.attrs[name])

    def __setitem__(self, name, value):
        """ set an item value
        """
        self.context.attrs[name] = value

    def __contains__(self, name):
        """ test if item exists
        """
        return name in self.context.attrs

    def __delitem__(self, name):
        """ remove an item
        """
        if name in self.context.attrs:
            del self.context.attrs[name]

    def uid(self):
        """ return internal identifier
        """
        return self.context.intid

    def __iter__(self):
        """ return all items
        """
        return iter(self.context.attrs)

    def items(self):
        """ return all items
        """
        return dict(self.context.attrs)
